/*   implementation of the own backend to use UdmSearch without third-party
 *   libraries
 */

#include "udm_config.h"

#ifdef HAVE_FILES
#include <stdio.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "udm_common.h"
#include "udm_spell.h"
#include "udm_db.h"
#include "udm_unicode.h"
#include "udm_robots.h"
#include "udm_mutex.h"
#include "udm_utils.h"
#include "udm_log.h"
#include "udm_conf.h"
#include "udm_crc32.h"
#include "udm_cache.h"
#include "udm_server.h"
#include "udm_xmalloc.h"
#include "udm_searchtool.h"
#include "udm_boolean.h"
#include "udm_stopwords.h"
#include "udm_doc.h"

typedef struct {
	int	url_id;
	uint4	coord;
	int	wrd_id;
} UDM_CRCWORD;



/************************* Backen functions *****************/
#define MAXMULTI	32
typedef struct txt_db {
	int connected;
	FILE * dict;
	FILE * url;
	int  crcdict[MAXMULTI];
	int  open_mode;
	int  errcode;
	char errmsg[1024];
} DB;

static void CloseDB(void * vdb){
	DB * db;
	int i;
	db=(DB*)vdb;
	if(db->dict)fclose(db->dict);
	if(db->url)fclose(db->url);
	for(i=0;i<MAXMULTI;i++)
		if(db->crcdict[i]>=0)
			close(db->crcdict[i]);
}

void UdmFreeDB(void * vdb){
	DB * db=(DB*)vdb;
        if(db){                                                        
                CloseDB(db);
                free(db);
        }                                                                       
}

static int InitDB(UDM_AGENT * Indexer){
	const char * fmode;
	int  flags,perm=0;
	int  i;
	DB * db=Indexer->db;
	
	if(db->connected)
		return(0);
	else
		db->connected=1;

	for(i=0;i<MAXMULTI;i++)
		db->crcdict[i]=-1;
	switch(db->open_mode){
		case UDM_OPEN_MODE_READ:
			fmode="r";
			flags=O_RDONLY|UDM_BINARY;
			break;
		case UDM_OPEN_MODE_WRITE:
		default:
			fmode="w";
			flags=O_WRONLY|O_CREAT|O_TRUNC|UDM_BINARY;
			perm=UDM_IWRITE;
	}

	if(Indexer->Conf->DBMode==UDM_DBMODE_SINGLE_CRC){
		char fname[1024]="";
		snprintf(fname,sizeof(fname)-1,"%s%s",Indexer->Conf->vardir,"dict");
		db->crcdict[0]=open(fname,flags,perm);
		if(db->crcdict[0]<0){
			sprintf(db->errmsg,"Can't open dict file '%s' (%s)", fname, strerror(errno));
			db->errcode=1;
			return(1);
		}
		sprintf(fname,"%s%s",Indexer->Conf->vardir,"url.txt");
		db->url=fopen(fname,fmode);
		if(!db->url){
			sprintf(db->errmsg,"Can't open URL file '%s' (%s)", fname, strerror(errno));
			db->errcode=1;
			return(1);
		}
	}else
	if(Indexer->Conf->DBMode==UDM_DBMODE_MULTI_CRC){
		int fnum;
		char fname[1024];
		for(fnum=1;fnum<MAXMULTI;fnum++){	
			snprintf(fname,sizeof(fname)-1,"%sdict%02d",Indexer->Conf->vardir,fnum);
			db->crcdict[fnum]=open(fname,flags,perm);
			if(db->crcdict[fnum]<0){
				sprintf(db->errmsg,"Can't open dict file '%s' (%s)", fname, strerror(errno));
				db->errcode=1;
				return(1);
			}
		}
		sprintf(fname,"%s%s",Indexer->Conf->vardir,"url.txt");
		db->url=fopen(fname,fmode);
		if(!db->url){
			sprintf(db->errmsg,"Can't open URL file '%s' (%s)", fname, strerror(errno));
			db->errcode=1;
			return(1);
		}
	}else{
		char fname[1024];
		snprintf(fname,sizeof(fname)-1,"%s%s",Indexer->Conf->vardir,"dict.txt");
		db->dict=fopen(fname,fmode);
		if(!db->dict){
			sprintf(db->errmsg,"Can't open dict file '%s' (%s)", fname, strerror(errno));
			db->errcode=1;
			return(1);
		}
		sprintf(fname,"%s%s",Indexer->Conf->vardir,"url.txt");
		db->url=fopen(fname,fmode);
		if(!db->url){
			sprintf(db->errmsg,"Can't open URL file '%s' (%s)", fname, strerror(errno));
			db->errcode=1;
			return(1);
		}
	}
	return(0);
}

void * UdmAllocDB(UDM_AGENT * Indexer, const char * DBAddr, int mode){
	DB * db;
	db=(DB*)malloc(sizeof(DB));
	bzero(db,sizeof(*db));
	db->open_mode=mode;
	return((void*)db);
}
char * UdmDBErrorMsg(void *db){
	return(((DB*)db)->errmsg);
}
int UdmDBErrorCode(void *db){
	return(((DB*)db)->errcode);
}
void UdmSetDBErrorCode(void * db,int code){
	((DB*)db)->errcode=code;
}

/*************  Words ****************************/

static int UdmDeleteWordFromURL(UDM_AGENT* Indexer,UDM_DOCUMENT * Doc){
	if(Indexer->Conf->DBMode==UDM_DBMODE_CACHE){
		int i;
		i=UdmDeleteURLFromCache(Indexer,Doc->url_id);
		return(i);
	}
	return(IND_OK);
}
static int UdmDeleteCrossWordFromURL(UDM_AGENT* Indexer,UDM_DOCUMENT * Doc){
	return(IND_OK);
}
static int UdmStoreCrossWords(UDM_AGENT * Indexer,UDM_DOCUMENT * Doc){
	return IND_OK;
}
static int UdmStoreWords(UDM_AGENT * Indexer, UDM_DOCUMENT * Doc){
	if (InitDB(Indexer)){
		UdmLog(Indexer, UDM_LOG_ERROR, "Error: %s", ((DB*)(Indexer->db))->errmsg);
		exit(1);
	}
	if(Indexer->Conf->DBMode==UDM_DBMODE_SINGLE_CRC){
		int fd,i;
		size_t bytes;
		UDM_CRCWORD * cw;
		if(Doc->WordList.nwords){
			fd=((DB*)(Indexer->db))->crcdict[0];
			bytes=sizeof(UDM_CRCWORD)*Doc->WordList.nwords;
			cw=(UDM_CRCWORD*)UdmXmalloc(bytes);
			for(i=0;i<Doc->WordList.nwords;i++){
				if(Doc->WordList.Word[i].coord){
					cw[i].url_id=Doc->url_id;
					cw[i].coord=Doc->WordList.Word[i].coord;
					cw[i].wrd_id=UdmStrCRC32(Doc->WordList.Word[i].word);
				}
			}
			if((bytes!=write(fd,(void *)cw,bytes))){
				sprintf(((DB*)(Indexer->db))->errmsg,"Can't write to dict file (%s)", strerror(errno));
				((DB*)(Indexer->db))->errcode=1;
				return(IND_ERROR);
			}
			free(cw);
		}
	}else
	if(Indexer->Conf->DBMode==UDM_DBMODE_MULTI_CRC){
		int fd,i,j,len;
		size_t bytes;
		UDM_CRCWORD * cw;
		if(Doc->WordList.nwords){
			bytes=sizeof(UDM_CRCWORD)*Doc->WordList.nwords;
			cw=(UDM_CRCWORD*)UdmXmalloc(bytes);
			for(len=1;len<MAXMULTI;len++){
				j=0;
				for(i=0;i<Doc->WordList.nwords;i++)
				if((Doc->WordList.Word[i].coord)&&len==strlen(Doc->WordList.Word[i].word)){
					cw[j].url_id=Doc->url_id;
					cw[j].coord=Doc->WordList.Word[i].coord;
					cw[j].wrd_id=UdmStrCRC32(Doc->WordList.Word[i].word);
					j++;
				}
				fd=((DB*)(Indexer->db))->crcdict[len];
				bytes=sizeof(UDM_CRCWORD)*j;
				if((bytes!=write(fd,(void *)cw,bytes))){
					sprintf(((DB*)(Indexer->db))->errmsg,"Can't write to dict file (%s)", strerror(errno));
					((DB*)(Indexer->db))->errcode=1;
					return(IND_ERROR);
				}
			}
			free(cw);
		}
	}else
	if(Indexer->Conf->DBMode){
		int res;
		res=UdmStoreWordsCache(Indexer,Doc);
	}else{
		FILE *f;
		int i;
		f=((DB*)(Indexer->db))->dict;
		for(i=0;i<Doc->WordList.nwords;i++){
			if(Doc->WordList.Word[i].coord){
				fprintf(f,"%d\t%d\t%s\n",
					Doc->url_id,
					Doc->WordList.Word[i].coord,
					Doc->WordList.Word[i].word);
			}
		}
	}
	return(IND_OK);
}

/************ URLs stuff *******************************/

static int n_urls=0;
static UDM_DOCUMENT* URLs=NULL;


UDM_DOCUMENT * UdmGetDocInfo(UDM_AGENT* Indexer,int expired_first){
	int i;
	UDM_DOCUMENT * Result;

	for(i=0;i<n_urls;i++){
		if(!URLs[i].status){
			URLs[i].status=1;
			Result=UdmAllocDoc();
			Result->url=strdup(URLs[i].url);
			Result->url_id=URLs[i].url_id;
			Result->hops=URLs[i].hops;
			return(Result);
		}
	}
	return(NULL);
}
static int UdmFindURL(UDM_AGENT* Indexer,UDM_DOCUMENT * Doc){
	return IND_OK;
}

static int UdmAddURL(UDM_AGENT* Indexer,UDM_DOCUMENT * Doc){
	URLs=(UDM_DOCUMENT *)UdmXrealloc(URLs,sizeof(UDM_DOCUMENT)*(n_urls+1));
	UdmInitDoc(&URLs[n_urls]);
	URLs[n_urls].url_id=n_urls+1;
	URLs[n_urls].status=0;
	URLs[n_urls].hops=Doc->hops;
	URLs[n_urls].referrer=Doc->referrer;
	URLs[n_urls].url=strdup(Doc->url);
	n_urls++;
	return(IND_OK);
}
static int UdmDeleteURL(UDM_AGENT* Indexer,UDM_DOCUMENT * Doc){
	return(IND_OK);
}
static int UdmUpdateUrl(UDM_AGENT* Indexer,UDM_DOCUMENT * Doc){
	return(IND_OK);
}

#define XVAL(x)		(!x?"-":!x[0]?"-":x)
static int UdmLongUpdateURL(UDM_AGENT* Indexer,UDM_DOCUMENT * Doc){
	int i;
	FILE *f;

	if (InitDB(Indexer)){
		UdmLog(Indexer, UDM_LOG_ERROR, "Error: %s", ((DB*)(Indexer->db))->errmsg);
		exit(1);
	}

	f=((DB*)(Indexer->db))->url;

	for(i=0;i<n_urls;i++)
	if(URLs[i].url_id==Doc->url_id){

		fprintf(f,"%d\t%s\t%d\t%d\t%s\t%s\t%d\t%s\t%li\t%s\t%s\t%s\t%s\n",
			Doc->url_id,
			XVAL(Doc->tag),
			Doc->status,
			Doc->size,
			XVAL(Doc->content_type),
			XVAL(Doc->lang),
			Doc->crc32,
			XVAL(URLs[i].url),
			Doc->last_mod_time,
			XVAL(Doc->text),
			XVAL(Doc->title),
			XVAL(Doc->keywords),
			XVAL(Doc->description));
		break;
	}
	return(IND_OK);
}



/************* Clones ***************************/
static int UdmFindOrigin(UDM_AGENT* Indexer, UDM_DOCUMENT * Doc){
	return(IND_OK);
}

static int UdmUpdateClone(UDM_AGENT* Indexer,UDM_DOCUMENT * Doc){
	return(IND_OK);
}

UDM_RESULT * UdmCloneList(UDM_AGENT * Indexer, UDM_DOCUMENT * Doc){
	return(NULL);
}

static int UdmRegisterChild(UDM_AGENT *Indexer, UDM_DOCUMENT * Doc){
	return IND_OK;
}
static int UdmFindMessage(UDM_AGENT *Indexer, UDM_DOCUMENT * Doc){
	return IND_OK;
}


/************* Clearing **************************/

int UdmClearDB(UDM_AGENT* Indexer){
	return(IND_OK);
}

/*************** Servers *************************/

int UdmLoadServerTable(UDM_AGENT * Indexer, char * name, int flags){
	return(IND_OK);
}

/**************** search stuff *******************/

static UDM_URL_CRD * UdmFindBuiltIn(UDM_AGENT * query, const char * text){
	int i,j,url_id;
	UDM_URL_CRD * wrd=NULL;
	char str[4048];

	UdmPrepare(query,text);

	/* Now find each word */
	if(query->Conf->DBMode==UDM_DBMODE_SINGLE_CRC){
		UDM_CRCWORD cw[256];
		int bytes, wnum;
		while((bytes=read(((DB*)(query->db))->crcdict[0],&cw,sizeof(cw)))){
			wnum=bytes/sizeof(UDM_CRCWORD);
			for(j=0;j<wnum;j++){
				for(i=0;i<query->WWList.nwords;i++){
					if(query->WWList.Words[i].crcword==cw[j].wrd_id){
						uint4 section;
						uint4 coord;
						uint4 weight;
						
						coord=cw[j].coord;
						section=UDM_WRDSEC(coord);
						weight=query->wf[section];
						
						if(weight){
							if(!query->total_found){
								wrd=(UDM_URL_CRD*)UdmXmalloc((query->total_found+1)*sizeof(UDM_URL_CRD));
							}else{
								wrd=(UDM_URL_CRD*)UdmXrealloc(wrd,(query->total_found+1)*sizeof(UDM_URL_CRD));
							}
							coord=coord&0xFFFF0000;
							wrd[query->total_found].url_id=cw[j].url_id;
							wrd[query->total_found].coord=coord+(weight<<8)+query->WWList.Words[i].order;
							query->total_found++;
						}
					}
				}
			}
		}
	}else
	if(query->Conf->DBMode==UDM_DBMODE_MULTI_CRC){
		UDM_CRCWORD cw[256];
		int bytes,len,fd,wnum;

		for(i=0;i<query->WWList.nwords;i++){
			len=strlen(query->WWList.Words[i].word);
			if(len>=MAXMULTI)len=MAXMULTI-1;
			fd=((DB*)(query->db))->crcdict[len];
			lseek(fd,(off_t)0,SEEK_SET);
			while((bytes=read(fd,&cw,sizeof(cw)))){
				wnum=bytes/sizeof(UDM_CRCWORD);
				for(j=0;j<wnum;j++){
					if(query->WWList.Words[i].crcword==cw[j].wrd_id){
						uint4 section;
						uint4 coord;
						uint4 weight;
						
						coord=cw[j].coord;
						section=UDM_WRDSEC(coord);
						weight=query->wf[section];
						
						if(weight){
							if(!query->total_found){
								wrd=(UDM_URL_CRD*)UdmXmalloc((query->total_found+1)*sizeof(UDM_URL_CRD));
							}else{
								wrd=(UDM_URL_CRD*)UdmXrealloc(wrd,(query->total_found+1)*sizeof(UDM_URL_CRD));
							}
							coord=coord&0xFFFF0000;
							wrd[query->total_found].url_id=cw[j].url_id;
							wrd[query->total_found].coord=coord+(weight<<8)+query->WWList.Words[i].order;
							query->total_found++;
						}
					}
				}
			}
		}
	}else{
		/* UDM_DBMODE_SINGLE */
		while(fgets(str,sizeof(str),((DB*)(query->db))->dict)){
			char * w, * lasttok=NULL;
			uint4 coord;
						
			w=UdmGetToken(str,"\t",&lasttok);
			if(!w)continue;url_id=atoi(w);
			w=UdmGetToken(NULL,"\t",&lasttok);
			if(!w)continue;coord=atoi(w);
			w=UdmGetToken(NULL,"\t\n",&lasttok);
			if(!w)continue;
			
			for(i=0;i<query->WWList.nwords;i++){
				int matches=0;
				
				switch(query->word_match){
					case UDM_MATCH_BEGIN:
						matches=!strncmp(w,query->WWList.Words[i].word,strlen(query->WWList.Words[i].word));
						break;
					case UDM_MATCH_END:
						if(strlen(query->WWList.Words[i].word)<=strlen(w)){
							char *ending;
							ending=w+strlen(w)-strlen(query->WWList.Words[i].word);
							matches=!strcmp(ending,query->WWList.Words[i].word);
						}
						break;
					case UDM_MATCH_SUBSTR:
						matches=(strstr(w,query->WWList.Words[i].word)!=NULL);
						break;
					case UDM_MATCH_FULL:
					default:
						matches=!strcmp(w,query->WWList.Words[i].word);
						break;
				}
				
				if(matches){
					uint4 section;
					uint4 weight;
					
					section=UDM_WRDSEC(coord);
					weight=query->wf[section];
					
					if(weight){
						if(!query->total_found){
							wrd=(UDM_URL_CRD*)malloc((query->total_found+1)*sizeof(UDM_URL_CRD));
						}else{
							wrd=(UDM_URL_CRD*)realloc(wrd,(query->total_found+1)*sizeof(UDM_URL_CRD));
						}
						
						wrd[query->total_found].url_id=url_id;
						wrd[query->total_found].coord=coord+(weight<<8)+query->WWList.Words[i].order;
						query->total_found++;
					}
				}
			}
		}
	}

	return wrd;
}

int UdmResAddDocInfo(UDM_AGENT * query, UDM_RESULT * Res){
	return IND_OK;
}

UDM_URL_CRD * UdmFindWords(UDM_AGENT * query, const char * words){
	return NULL;
}

UDM_RESULT * UdmFind(UDM_AGENT * query, const char * text){
	char str[1024];
	int url_id;
	int i,j;
	UDM_URL_CRD * wrd=NULL;
	UDM_RESULT * Res;
	int first;
	int skip=0;
	unsigned long ticks;

	if(InitDB(query))return(NULL);
	ticks=UdmStartTimer();

	if(UdmDBErrorCode(query->db))return(NULL);

	if(query->Conf->DBMode==UDM_DBMODE_CACHE){
		wrd=UdmFindCache(query,text);
	}else{
		wrd=UdmFindBuiltIn(query,text);
	}

	/* Allocate result */
	Res=(UDM_RESULT*)malloc(sizeof(UDM_RESULT));
	Res->Doc=NULL;
	Res->total_found=0;
	Res->num_rows=0;
	strcpy(Res->wordinfo,query->wordinfo);

	/* Return if nothing was found */
	if(!query->total_found){
		Res->work_time=UdmStartTimer()-ticks;
		return(Res);
	}

	/* Now let's sort in url_id order then group results */
	if(query->total_found){
		UdmSortSearchWordsByURL(wrd,query->total_found);
		UdmGroupByURL(query,wrd);
	}

	/* Sort by the weight */
	if(query->page_number<FAST_PRESORT_PAGES){
		size_t topcount;
		topcount=query->page_size*(query->page_number+1)-1;
		if(topcount>=query->total_found)topcount=query->total_found-1;
		UdmWrdTopSort(wrd,query->total_found,topcount);
	}else{
		UdmSortSearchWordsByWeight(wrd,query->total_found);
	}

	/* Copy SEARCHWORD to DOC structure */
	Res->Doc=(UDM_DOCUMENT*)UdmXmalloc(sizeof(UDM_DOCUMENT)*(query->total_found));
	for(i=0;i<query->total_found;i++){
		UdmInitDoc(&Res->Doc[i]);
		Res->Doc[i].url_id=wrd[i].url_id;
		Res->Doc[i].rating=123; /* FIXME: word count?? */
	}
	free(wrd);

	/* Now apply URL limits */
	while(fgets(str,sizeof(str),((DB*)(query->db))->url)){
		url_id=atoi(str);
		for(i=0;i<query->total_found-skip;i++){
			if(url_id==Res->Doc[i].url_id){
				char * tag=NULL, *w, *lasttok=NULL;
				int matched=1;
				w=UdmGetToken(str,"\t",&lasttok);
				for(j=0;w;j++){
					switch(j){
						case  1: tag=strdup(w);break;
						case  3: Res->Doc[i].size=atoi(w);break;
						case  4: Res->Doc[i].content_type=strdup(w);break;
						case  6: Res->Doc[i].crc32=atoi(w);break;
						case  7: Res->Doc[i].url=strdup(w);break;
						case  8: Res->Doc[i].last_mod_time=atol(w);break;
						case  9: Res->Doc[i].text=strdup(w);break;
						case 10: Res->Doc[i].title=strdup(w);break;
						case 11: Res->Doc[i].keywords=strdup(w);break;
						case 12: Res->Doc[i].description=strdup(w);break;
					}
					
					w=UdmGetToken(NULL,"\t",&lasttok);
					
				}

				/* Check "ul" subsection match  */
				if(query->Conf->ul)matched=(strstr(Res->Doc[i].url,query->Conf->ul)!=NULL);
				/* Check "tag" subsection match */
				if((matched)&&(query->Conf->ttag)&&(tag))matched=(!strcmp(query->Conf->ttag,tag));

				if(!matched){
					/* Skip this URL */
					UDM_FREE(Res->Doc[i].content_type);
					UDM_FREE(Res->Doc[i].url);
					UDM_FREE(Res->Doc[i].text);
					UDM_FREE(Res->Doc[i].title);
					UDM_FREE(Res->Doc[i].keywords);
					UDM_FREE(Res->Doc[i].description);
					skip++;
					memmove(&(Res->Doc[i]),&(Res->Doc[i+1]),(query->total_found-i-1)*sizeof(UDM_DOCUMENT));
				}
			}
		}
	}
	
	Res->total_found =query->total_found-skip;
	for(i=0;i<Res->total_found;i++){
		Res->Doc[i].order=i+1;
	}
	first=query->page_number*query->page_size;
	if(first>query->total_found-skip)first=query->total_found-skip;

	Res->num_rows = query->total_found-skip-first;
	if((query->total_found-skip-first)>query->page_size)
	{
		Res->num_rows = query->page_size;
	}
	if(first>0){
		for(i=0;i<first;i++){
			UDM_FREE(Res->Doc[i].content_type);
			UDM_FREE(Res->Doc[i].url);
			UDM_FREE(Res->Doc[i].text);
			UDM_FREE(Res->Doc[i].title);
			UDM_FREE(Res->Doc[i].keywords);
			UDM_FREE(Res->Doc[i].description);
		}
		memmove(&(Res->Doc[0]),&(Res->Doc[first]),Res->num_rows*sizeof(UDM_DOCUMENT));
		/* FIXME realloc required */
	}
	Res->work_time=UdmStartTimer()-ticks;
	Res->first=first+1;
	Res->last=Res->first+Res->num_rows-1;
	query->total_found=query->total_found-skip;
	query->total_found=Res->total_found;
	return(Res);
}



/***************** Misc functions ***************/


UDM_CATEGORY * UdmCatPath(UDM_AGENT * Indexer,const char * category){
	return(NULL);
}

UDM_CATEGORY * UdmCatList(UDM_AGENT * Indexer,const char * category){
	return(NULL);
}

__INDLIB__ int  UdmGetStatistics(UDM_AGENT * Indexer){
	if(!Indexer->Conf->StatInfo)return(IND_OK);
	Indexer->Conf->StatInfo(Indexer->handle,0,0,0,"Stats not implemented in built-in database");
	Indexer->Conf->StatInfo(Indexer->handle,-1,0,0,"");
	return(IND_OK);
}
__INDLIB__ int UdmMarkForReindex(UDM_AGENT* Indexer){
	return(IND_OK);
}
__INDLIB__ int UdmAddTagLimit(UDM_ENV * Conf,char * tag){
	Conf->ttag=strdup(tag);
	return(0);
}
__INDLIB__ int UdmAddStatusLimit(UDM_ENV * Conf,int status){
	return(0);
}
__INDLIB__ int UdmAddURLLimit(UDM_ENV * Conf,char *URL){
	UDM_FREE(Conf->ul);
	Conf->ul=strdup(URL);
	return(0);
}
__INDLIB__ int UdmAddLangLimit(UDM_ENV * Conf,char *URL){
	return(0);
}

__INDLIB__ int UdmAddCatLimit(UDM_ENV * Conf,char * cat){
	return(0);
}

__INDLIB__ int UdmClearURLLimit(UDM_ENV * Conf){
	UDM_FREE(Conf->ul);
	UDM_FREE(Conf->ttag);
	return(0);
}
__INDLIB__ int UdmClearLimits(UDM_ENV * Conf){
	UDM_FREE(Conf->ul)
	return(0);
}
__INDLIB__ int UdmGetReferers(UDM_AGENT* Indexer){
	return(IND_OK);
}
__INDLIB__ int UdmAddTimeLimit(UDM_ENV * Conf,struct udm_stl_info_t * stl){
 return 0;
}
__INDLIB__ int UdmGetDocCount(UDM_AGENT * Indexer){
	return 0;
}


__INDLIB__ UDM_URLSTATE * UdmGetURLState(UDM_AGENT * Indexer, const char * hostinfo,time_t lmt){
	return NULL;
}


UDM_UINT8_URLID* GetPair8FromSQL(UDM_AGENT *Agent,const char *field,int *count,int type){
	return 0;
}
UDM_UINT4_URLID* GetPair4FromSQL(UDM_AGENT *Agent,const char *field,int *count,int type){
	return 0;
}



int UdmURLAction(UDM_AGENT * A, UDM_DOCUMENT * D, int cmd){
	int res;

	switch(cmd){
		case UDM_URL_ACTION_DELETE:
			res=UdmDeleteURL(A,D);
			break;
			
		case UDM_URL_ACTION_ADD:
			res=UdmAddURL(A,D);
			break;
			
		case UDM_URL_ACTION_SUPDATE:
			res=UdmUpdateUrl(A,D);
			break;
			
		case UDM_URL_ACTION_LUPDATE:
			res=UdmLongUpdateURL(A,D);
			break;
			
		case UDM_URL_ACTION_INSWORDS:
			res=UdmStoreWords(A,D);
			break;
			
		case UDM_URL_ACTION_INSCWORDS:
			res=UdmStoreCrossWords(A,D);
			break;
			
		case UDM_URL_ACTION_DELWORDS:
			res=UdmDeleteWordFromURL(A,D);
			break;
			
		case UDM_URL_ACTION_DELCWORDS:
			res=UdmDeleteCrossWordFromURL(A,D);
			break;
			
		case UDM_URL_ACTION_UPDCLONE:
			res=UdmUpdateClone(A,D);
			break;
			
		case UDM_URL_ACTION_REGCHILD:
			res=UdmRegisterChild(A,D);
			break;
			
		case UDM_URL_ACTION_FINDBYURL:
			res=UdmFindURL(A,D);
			break;
			
		case UDM_URL_ACTION_FINDBYMSG:
			res=UdmFindMessage(A,D);
			break;
			
		case UDM_URL_ACTION_FINDORIG:
			res=UdmFindOrigin(A,D);
			break;
			
		default:
			res=IND_ERROR;
	}
	return res;
}


#endif
